"""Support code for production hardware tests

We execute the production (hardware tests) for new hardware. Usually this
will be one of the following hardware units:

- stationary transceiver unit
- sensory tool holder (STH)/sensory tool assembly
- sensory milling head (SMH)
"""
